/* Copyright 2016 Dirk Duschinger */

//////////////////////////////////////
// author : Dirk Duschinger         //
// email  : dirk.duschigner@cern.ch //
//////////////////////////////////////

#include <string>
#include <algorithm>
#include <vector>

#include "EpsPolygon.h"
#include <iterator>

EpsPolygon::EpsPolygon()
  : m_dRotation(0)
  , m_bFill(false) {
}

EpsPolygon::~EpsPolygon() {
}

void EpsPolygon::Draw(WINDOW* w, int eps_xres, int eps_yres) {
  if (m_bFill) {
    DrawFill(w, eps_xres, eps_yres);
    return;
  }
  if (m_sTitle != "") {
    DrawString(w, eps_xres, eps_yres);
    return;
  }
  if (GetCoordinateList().size() > 1) {
    // std::vector<std::vector<double> >::iterator
    auto it = GetCoordinateList().begin();
    for (; std::next(it) != GetCoordinateList().end(); ++it) {
      EpsLine* l = new EpsLine((*it).at(0),
                               (*it).at(1),
                               std::next(it)->at(0),
                               std::next(it)->at(1),
                               GetColor());
      l->Draw(w, eps_xres, eps_yres);
    }
  }
}

void EpsPolygon::DrawString(WINDOW* w, int eps_xres, int eps_yres) {
  int curses_xres, curses_yres;
  getmaxyx(w, curses_yres, curses_xres);
  double xscale = 1./(1.*eps_xres/(curses_xres-1));
  double yscale = 1./(1.*eps_yres/(curses_yres-1));
  double x = 1. * GetCoordinateList().begin()->at(0) * xscale;
  double y = 1. * GetCoordinateList().begin()->at(1) * yscale;
  correct_string(&(m_sTitle));
  int s_length = m_sTitle.length();
  double slope = m_dRotation;
  // flip vertically
  y = curses_yres - y -1;
  for (int i=0; i != s_length; i++) {
    int cx = static_cast<int>(y+i*sin(slope));
    int cy = static_cast<int>(x+i*cos(slope));
    int col = GetColor();
    int fg = col%8;
    int bg = (mvwinch(w, cx, cy) & A_COLOR) >> 11;

    wcolor_set(w, fg+bg*8, 0);
    mvwaddch(w, cx, cy, chtype(m_sTitle.at(i)));
    wcolor_set(w, 0, 0);
  }
}

void EpsPolygon::DrawFill(WINDOW* w, int eps_xres, int eps_yres) {
  double xmin, ymin, xmax, ymax;
  std::vector<std::vector<double> >::iterator it = GetCoordinateList().begin();
  for (; std::next(it) != GetCoordinateList().end(); ++it) {
    if (GetCoordinateList().size() < 2)
      break;
    if (it == GetCoordinateList().begin()) {
      xmin = it->at(0);
      ymin = it->at(1);
      xmax = it->at(0);
      ymax = it->at(1);
    } else {
      xmin = std::min(xmin, it->at(0));
      ymin = std::min(ymin, it->at(1));
      xmax = std::max(xmax, it->at(0));
      ymax = std::max(ymax, it->at(1));
    }
  }
  int curses_xres, curses_yres;
  getmaxyx(w, curses_yres, curses_xres);
  double xscale = 1./(1.*eps_xres/(curses_xres-1));
  double yscale = 1./(1.*eps_yres/(curses_yres-1));

  double x1 = 1. * xmin * xscale;
  double y1 = 1. * ymin * yscale;
  double x2 = 1. * xmax * xscale;
  double y2 = 1. * ymax * yscale;
  y1 = curses_yres - y1 -1;
  y2 = curses_yres - y2 -1;

  for (int x=static_cast<int>(x1); x<static_cast<int>(x2); x++) {
    for (int y=static_cast<int>(y2); y<static_cast<int>(y1); y++) {
      if (testInsidePolygon(x,
                            y,
                            eps_xres,
                            eps_yres) == true) {
        wcolor_set(w, GetColor(), 0);
        mvwaddch(w, y, x, ' ');
        wcolor_set(w, 0, 0);
      }
    }
  }
}

bool EpsPolygon::testInsidePolygon(int x, int y, int eps_xres, int eps_yres) {
  int t = -1;
  std::vector<std::vector<double> >::iterator it = GetCoordinateList().begin();
  for (; std::next(it) != GetCoordinateList().end(); ++it) {
    t *= testCrossLines(x,
                        y,
                        it->at(0),
                        it->at(1),
                        next(it)->at(0),
                        next(it)->at(1),
                        eps_xres,
                        eps_yres);
  }
  return (t != -1);
}

int EpsPolygon::testCrossLines(int x,
                               int y,
                               int x1,
                               int y1,
                               int x2,
                               int y2,
                               int eps_xres,
                               int eps_yres) {
  int curses_xres, curses_yres;
  getmaxyx(stdscr, curses_yres, curses_xres);
  double xscale = 1./(1.*eps_xres/(curses_xres-1));
  double yscale = 1./(1.*eps_yres/(curses_yres-1));
  x1 *= xscale;
  y1 *= yscale;
  x2 *= xscale;
  y2 *= yscale;

  y1 = curses_yres - y1 -1;
  y2 = curses_yres - y2 -1;
  if ((y == y1) && (y == y2)) {
    if (((x1 <= x) && (x <= x2)) || ((x2 <= x) && (x <= x1)))
      return 0;
    else
      return 1;
  }
  if (y1 > y2) {
    std::swap(y1, y2);
  }
  if ((y == y1) && (x == x1))
    return 0;
  if ((y <= y1) || (y > y2))
    return 1;
  int delta = (x1-x)*(y2-y)-(y1-y)*(x2-x);
  if (delta > 0)
    return -1;
  if (delta < 0)
    return 1;
  return 0;
}

// bool EpsPolygon::GetFill(){
//   return m_bFill;
// }

// double EpsPolygon::GetRotation(){
//   return m_dRotation;
// }

// std::string EpsPolygon::GetTitle(){
//   return m_sTitle;
// }

void EpsPolygon::SetFill(bool fill) {
  m_bFill = fill;
}

void EpsPolygon::SetRotation(double r) {
  m_dRotation = r* M_PI / 180;
}

void EpsPolygon::SetTitle(std::string s) {
  m_sTitle = s;
}


std::ostream& operator<<(std::ostream& os, EpsPolygon* ep) {
  os << "===========================";
  os << "\nPolygon: " << "\n";
  os << "Color: " << ep->GetColor();
  if (ep->GetCoordinateList().size() > 1) {
    std::vector<std::vector<double> >::iterator it = \
      ep->GetCoordinateList().begin();
    for (; std::next(it) != ep->GetCoordinateList().end(); ++it) {
      os << "\nCoordinate " << distance(ep->GetCoordinateList().begin(), it) \
         << ": " << it->at(0) << ", " << it->at(1);
    }
  }
  os << "\n===========================";
  return os;
}

void correct_string(std::string* s) {
  int pos;
  pos = s->find("\\255");
  if (pos != -1)
    s->replace(pos, 4, "-");
  pos = s->find("\\)");
  if (pos != -1)
    s->replace(pos, 2, ")");
  pos = s->find("\\(");
  if (pos != -1)
    s->replace(pos, 2, "(");
}
